import pymysql
from twisted.internet.threads import deferToThread


class MySQLPipeline:
    def __init__(self, host, database, user, password, port):
        self.host = host
        self.database = database
        self.user = user
        self.password = password
        self.port = port

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            host=crawler.settings.get('MYSQL_HOST'),
            database=crawler.settings.get('MYSQL_DATABASE'),
            user=crawler.settings.get('MYSQL_USER'),
            password=crawler.settings.get('MYSQL_PASSWORD'),
            port=crawler.settings.get('MYSQL_PORT'),
        )

    def open_spider(self, spider):
        self.db = pymysql.connect(self.host, self.user, self.password, self.database, charset='utf8',
                                  port=self.port)
        self.cursor = self.db.cursor()

    def close_spider(self, spider):
        self.db.close()

    def _has_dict_item(self, table, item):
        where_strs = f"`unique_identity` = %s and `{item.unique_key}` = %s"
        where_value_list = [item.unique_identity, item.get(item.unique_key)]
        query = 'select %s from %s where %s' % (item.unique_key, table, where_strs)

        self.cursor.execute(query, tuple(where_value_list))
        rows = self.cursor.fetchone()

        if rows:
            return True
        else:
            return False

    def _insert_dict_item(self, table, item):
        data = dict(item)
        keys = ', '.join(f'`{data.keys()}`')
        values = ', '.join(['%s'] * len(data))
        sql = 'insert into %s (%s) values (%s)' % (table, keys, values)
        self.cursor.execute(sql, tuple(data.values()))
        self.db.commit()

    def _update_dict_item(self, table, item):
        data = dict(item)
        set_strs = []
        set_values = []

        for param, value in data.items():
            set_strs.append(f'`{param}` = %s')
            set_values.append(value)

        where_strs = f"`unique_identity` = %s and `{item.unique_key}` = %s"
        where_value_list = [item.unique_identity, data.get(item.unique_key)]

        set_str = ', '.join(set_strs)
        sql = "update %s set %s where %s" % (table, set_str, where_strs)
        self.cursor.execute(sql, tuple(data.values()) + tuple(where_value_list))
        self.db.commit()

    def _process_item(self, item, spider):
        allowed_spiders = item.mysql_spiders
        allowed_tables = item.mysql_tables
        if allowed_spiders and spider.name in allowed_spiders:
            for allowed_table in allowed_tables:
                if item.unique_key and item.unique_identity:
                    if self._has_dict_item(allowed_table, item):
                        self._update_dict_item(allowed_table, item)
                    else:
                        self._insert_dict_item(allowed_table, item)
                else:
                    self._insert_dict_item(allowed_table, item)
        return item

    def process_item(self, item, spider):
        return deferToThread(self._process_item, item, spider)



import logging

from .parsers import Parser
from .text import (StopWords, StopWordsArabic, StopWordsChinese,
                   StopWordsKorean, StopWordsHindi, StopWordsJapanese, StopWordsThai)

log = logging.getLogger(__name__)


class Configuration(object):
    def __init__(self):
        """
        Modify any of these Article / Source properties
        TODO: Have a separate ArticleConfig and SourceConfig extend this!
        """
        self.MAX_TITLE = 200  # num of chars
        self.MAX_TEXT = 100000  # num of chars
        self.MAX_AUTHORS = 10  # num strings in list

        # Don't toggle this variable, done internally
        self.use_meta_language = True

        # You may keep the html of just the main article body
        self.keep_article_html = True

        # English is the fallback
        self._language = 'zh'

        # Unique stopword classes for oriental languages, don't toggle
        self.stopwords_class = self.get_stopwords_class(self._language)

        self.verbose = False  # for debugging

    def get_language(self):
        return self._language

    def del_language(self):
        raise Exception('wtf are you doing?')

    def set_language(self, language):
        """Language setting must be set in this method b/c non-occidental
        (western) languages require a separate stopwords class.
        """
        if not language or len(language) != 2:
            raise Exception("Your input language must be a 2 char language code, \
                for example: english-->en \n and german-->de")

        # If explicitly set language, don't use meta
        self.use_meta_language = False

        # Set oriental language stopword class
        self._language = language
        self.stopwords_class = self.get_stopwords_class(language)

    language = property(get_language, set_language,
                        del_language, "language prop")

    @staticmethod
    def get_stopwords_class(language):
        if language == 'ko':
            return StopWordsKorean
        elif language == 'hi':
            return StopWordsHindi
        elif language == 'zh':
            return StopWordsChinese
        # Persian and Arabic Share an alphabet
        # There is a persian parser https://github.com/sobhe/hazm, but nltk is likely sufficient
        elif language == 'ar' or language == 'fa':
            return StopWordsArabic
        elif language == 'ja':
            return StopWordsJapanese
        elif language == 'th':
            return StopWordsThai
        return StopWords

    @staticmethod
    def get_parser():
        return Parser


class ArticleConfiguration(Configuration):
    pass



import asyncio
import logging

import pyppeteer
import websockets
from pyppeteer.errors import TimeoutError
from scrapy.http import HtmlResponse

pyppeteer_level = logging.WARNING
logging.getLogger('websockets.protocol').setLevel(pyppeteer_level)
logging.getLogger('pyppeteer').setLevel(pyppeteer_level)


class PyppeteerMiddleware:
    def __init__(self, **args):
        """
        init logger, loop, browser
        :param args:
        """
        self.args = args
        self.loop = asyncio.get_event_loop()

    def __del__(self):
        """
        close loop
        :return:
        """
        self.loop.close()

    def render(self, url, retries=1, script=None, wait=0.3, scrolldown=False, sleep=0,
               timeout=8.0, keep_page=False):
        """
        render page with pyppeteer
        :param url: page url
        :param retries: max retry times
        :param script: js script to evaluate
        :param wait: number of seconds to wait before loading the page, preventing timeouts
        :param scrolldown: how many times to page down
        :param sleep: how many long to sleep after initial render
        :param timeout: the longest wait time, otherwise raise timeout error
        :param keep_page: keep page not to be closed, browser object needed
        :return: content, [result]
        """
        browser = self.loop.run_until_complete(pyppeteer.launch(headless=True,
                                                                handleSIGTERM=False,
                                                                handleSIGINT=False))

        # define async render
        async def async_render(inner_url, inner_script, inner_scrolldown, inner_sleep, inner_wait, inner_timeout,
                               inner_keep_page):
            try:
                # basic render
                page = await browser.newPage()
                await asyncio.sleep(inner_wait)
                response = await page.goto(inner_url, options={'timeout': int(inner_timeout * 1000)})
                if response.status != 200:
                    return None, None, response.status
                inner_result = None
                # evaluate with script
                if inner_script:
                    inner_result = await page.evaluate(inner_script)

                # scroll down for {scrolldown} times
                if inner_scrolldown:
                    for _ in range(inner_scrolldown):
                        await page._keyboard.down('PageDown')
                        await asyncio.sleep(inner_sleep)
                else:
                    await asyncio.sleep(inner_sleep)
                if inner_scrolldown:
                    await page._keyboard.up('PageDown')

                # get html of page
                page_content = await page.content()

                return page_content, inner_result, response.status
            except TimeoutError:
                return None, None, 500
            finally:
                # if keep page, do not close it
                if not inner_keep_page:
                    await page.close()

        content, result, status = [None] * 3

        # retry for {retries} times
        for i in range(retries):
            if not content:
                content, result, status = self.loop.run_until_complete(
                    async_render(inner_url=url, inner_script=script, inner_sleep=sleep, inner_wait=wait,
                                 inner_scrolldown=scrolldown, inner_timeout=timeout, inner_keep_page=keep_page))
            else:
                break
        self.loop.run_until_complete(browser.close())
        # if need to return js evaluation result
        return content, result, status

    def process_request(self, request, spider):
        """
        :param request: request object
        :param spider: spider object
        :return: HtmlResponse
        """
        if request.meta.get('render'):
            try:
                html, result, status = self.render(request.url, **self.args)
                return HtmlResponse(url=request.url, body=html, request=request, encoding='utf-8',
                                    status=status)
            except websockets.exceptions.ConnectionClosed:
                pass

    @classmethod
    def from_crawler(cls, crawler):
        return cls(**crawler.settings.get('PYPPETEER_ARGS', {}))



import logging
import requests
import json


class CookiesMiddleware():
    def __init__(self, cookies_url):
        self.logger = logging.getLogger(__name__)
        self.cookies_url = cookies_url
    
    def get_random_cookies(self):
        try:
            response = requests.get(self.cookies_url)
            if response.status_code == 200:
                cookies = json.loads(response.text)
                return cookies
        except requests.ConnectionError:
            return False
    
    def process_request(self, request, spider):
        self.logger.debug('Getting Cookies')
        cookies = self.get_random_cookies()
        if cookies:
            request.cookies = cookies
            self.logger.debug('Using Cookies {cookies}'.format(cookies=json.dumps(cookies)))
    
    @classmethod
    def from_crawler(cls, crawler):
        settings = crawler.settings
        return cls(
            cookies_url=settings.get('COOKIES_URL')
        )



import codecs
import os
import re

from . import settings


class FileHelper(object):
    @staticmethod
    def loadResourceFile(filename):
        if not os.path.isabs(filename):
            dirpath = os.path.abspath(os.path.dirname(__file__))
            path = os.path.join(dirpath, 'resources', filename)
        else:
            path = filename
        try:
            f = codecs.open(path, 'r', 'utf-8')
            content = f.read()
            f.close()
            return content
        except IOError:
            raise IOError("Couldn't open file %s" % path)


class StringSplitter(object):
    def __init__(self, pattern):
        self.pattern = re.compile(pattern)

    def split(self, string):
        if not string:
            return []
        return self.pattern.split(string)


class StringReplacement(object):
    def __init__(self, pattern, replaceWith):
        self.pattern = pattern
        self.replaceWith = replaceWith

    def replaceAll(self, string):
        if not string:
            return ''
        return string.replace(self.pattern, self.replaceWith)


class ReplaceSequence(object):
    def __init__(self):
        self.replacements = []

    def create(self, firstPattern, replaceWith=None):
        result = StringReplacement(firstPattern, replaceWith or '')
        self.replacements.append(result)
        return self

    def append(self, pattern, replaceWith=None):
        return self.create(pattern, replaceWith)

    def replaceAll(self, string):
        if not string:
            return ''

        mutatedString = string
        for rp in self.replacements:
            mutatedString = rp.replaceAll(mutatedString)
        return mutatedString


def get_available_languages():
    """Returns a list of available languages and their 2 char input codes
    """
    stopword_files = os.listdir(os.path.join(settings.STOPWORDS_DIR))
    two_dig_codes = [f.split('-')[1].split('.')[0] for f in stopword_files]
    for d in two_dig_codes:
        assert len(d) == 2
    two_dig_codes.sort()
    return two_dig_codes


def extend_config(config, config_items):
    """
    We are handling config value setting like this for a cleaner api.
    Users just need to pass in a named param to this source and we can
    dynamically generate a config object for it.
    """
    for key, val in list(config_items.items()):
        if hasattr(config, key):
            setattr(config, key, val)

    return config



import logging
import requests


class ProxyMiddleware():
    def __init__(self, proxy_url, proxy_fail_times):
        self.logger = logging.getLogger(__name__)
        self.proxy_url = proxy_url
        self.proxy_fail_times = proxy_fail_times
    
    def get_random_proxy(self):
        try:
            response = requests.get(self.proxy_url)
            if response.status_code == 200:
                proxy = response.text
                return proxy
        except requests.ConnectionError:
            return False
    
    def process_request(self, request, spider):
        if request.meta.get('retry_times') >= self.proxy_fail_times:
            proxy = self.get_random_proxy()
            if proxy:
                uri = 'https://{proxy}'.format(proxy=proxy)
                self.logger.debug('Using Proxy {proxy}'.format(proxy=proxy))
                request.meta['proxy'] = uri
    
    @classmethod
    def from_crawler(cls, crawler):
        settings = crawler.settings
        return cls(
            proxy_url=settings.get('PROXY_URL'),
            proxy_fail_times=settings.get('PROXY_FAIL_TIMES', 1),
        )



import time

from furl import furl
from scrapy import Request
from scrapy.http import HtmlResponse
from scrapy.spiders import CrawlSpider as BaseSpider, signals
from scrapy.spiders.crawl import Rule as BaseRule
from scrapy.utils.request import request_fingerprint
from scrapy_helper.core.fast_item import FastItem

from scrapy_helper.core.urls import get_domain
from scrapy_helper.core.utils import str2body, str2dict, str2list, convert_to_html, convert_to_text
from scrapy_helper.newspaper.article import Article


class Rule(BaseRule):
    def __init__(self, link_extractor, callback=None, cb_kwargs=None, follow=None, process_links=None,
                 process_request=None, errback=None,
                 method='GET', params=None, process_body=None, data=None, headers=None,
                 priority=0, dont_filter=False, cookies=None,
                 meta=None, dont_redirect=None, dont_retry=None,
                 handle_httpstatus_list=None, handle_httpstatus_all=None,
                 dont_cache=None, dont_obey_robotstxt=None,
                 download_timeout=None, max_retry_times=None, proxy=None, render=False, splash=None,
                 persist_filter=False, extract_article=False):
        super().__init__(link_extractor, callback, cb_kwargs, follow, process_links, process_request, errback)

        self.method = method
        self.params = str2dict(params)
        self.process_body = process_body
        self.data = str2body(data)
        self.headers = str2dict(headers)
        self.priority = priority
        self.dont_filter = dont_filter
        self.cookies = str2body(cookies)
        self.meta = str2dict(meta)
        self.dont_redirect = dont_redirect
        self.dont_retry = dont_retry
        self.handle_httpstatus_list = str2list(handle_httpstatus_list, lambda x: int(x))
        self.handle_httpstatus_all = handle_httpstatus_all
        self.dont_cache = dont_cache
        self.dont_obey_robotstxt = dont_obey_robotstxt
        self.download_timeout = download_timeout
        self.max_retry_times = max_retry_times
        self.proxy = proxy
        self.render = render
        self.splash = str2body(splash)
        self.persist_filter = persist_filter
        self.extract_article = extract_article

    def __str__(self):
        """
        object to str
        :return:
        """
        return str(self.__dict__.items())


class CrawlSpider(BaseSpider):
    name = None

    def start_requests(self):
        """
        override start requests
        :return:
        """
        self.crawler.signals.connect(self.make_start_requests, signal=signals.spider_idle)
        return []

    def make_start_requests(self):
        """
        make start requests
        :return:
        """
        for request in self.start():
            self.crawler.engine.slot.scheduler.enqueue_request(request)

    def start(self):
        """
        start requests
        :return:
        """
        for url in self.make_start_urls():
            yield Request(url)

    def make_start_urls(self):
        """
        get start urls
        :return:
        """
        return self.start_urls

    def _generate_request(self, rule_index, rule, link, response):
        """
        generate request by rule
        :param rule_index: rule index
        :param rule: rule object
        :param link: link object
        :return: new request object
        """
        url = furl(link.url).add(rule.params).url if rule.params else link.url

        # init request body
        body = None
        # process by method
        if rule.method.upper() == 'POST':
            # if process_body defined, use its result
            if callable(rule.process_body):
                body = rule.process_body(response)
            # if data defined in rule, use data
            if rule.data:
                body = rule.data

        meta = dict(**rule.meta)
        meta_items = ['dont_redirect', 'dont_retry', 'handle_httpstatus_list', 'handle_httpstatus_all',
                      'dont_cache', 'dont_obey_robotstxt', 'download_timeout', 'max_retry_times', 'proxy', 'render',
                      'splash', 'persist_filter', 'extract_article']
        meta_args = {meta_item: getattr(rule, meta_item) for meta_item in meta_items if
                     not getattr(rule, meta_item) is None}
        meta.update(**meta_args)
        meta.update(rule=rule_index, link_text=link.text)

        r = Request(
            url=url,
            method=rule.method,
            body=body,
            headers=rule.headers,
            cookies=rule.cookies,
            priority=rule.priority,
            dont_filter=rule.dont_filter,
            callback=self._callback,
            errback=self._errback,
            meta=meta,
        )

        return r

    @staticmethod
    def _get_request_fp(request):
        return request.meta.get('request_fp') or request_fingerprint(request)

    @staticmethod
    def convert_to_html(element):
        return convert_to_html(element)

    @staticmethod
    def convert_to_text(element):
        return convert_to_text(element)

    @staticmethod
    def join_to_html(texts):
        return convert_to_html(''.join(texts))

    @staticmethod
    def join_to_text(texts):
        return convert_to_text(''.join(texts))

    @staticmethod
    def fast_item(*args: str):
        _item = FastItem()
        _item.set(args)
        return _item

    def _callback(self, response):
        request_fp = self._get_request_fp(response.request)
        request_domain = get_domain(response.url)
        response.meta.update(request_fp=request_fp, request_domain=request_domain,
                             insert_time=time.strftime('%Y-%m-%d %H:%M:%S'))

        if response.meta['extract_article']:
            article = Article(url=response.url, html=response.body)
            article.parse()
            response.meta.update(
                article={"title": article.title, "text": article.text, "article_html": article.article_html,
                         "images": article.images, "movies": article.movies, "tags": article.tags,
                         "authors": article.authors, "top_img": article.top_img,
                         "meta_img": article.meta_img, "meta_keywords": article.meta_keywords,
                         "meta_description": article.meta_description, "meta_lang": article.meta_lang,
                         "publish_date": article.publish_date, "meta_favicon": article.meta_favicon,
                         "meta_site_name": article.meta_site_name, "meta_data": article.meta_data,
                         "canonical_link": article.canonical_link})

        rule = self._rules[response.meta['rule']]
        return self._parse_response(response, rule.callback, rule.cb_kwargs, rule.follow)

    def _requests_to_follow(self, response):
        if not isinstance(response, HtmlResponse):
            return
        seen = set()
        for rule_index, rule in enumerate(self._rules):
            links = [lnk for lnk in rule.link_extractor.extract_links(response)
                     if lnk not in seen]
            for link in rule.process_links(links):
                seen.add(link)
                # request = self._build_request(rule_index, link)
                request = self._generate_request(rule_index, rule, link, response)
                yield rule.process_request(request, response)



from platinum2 import generate_user_agent
from scrapy import signals


class UserAgentMiddleware:
    def __init__(self, user_agent='Scrapy'):
        self.user_agent = user_agent

    @classmethod
    def from_crawler(cls, crawler):
        cls.mode = crawler.settings.get('USER_AGENT_MODE')
        cls.os = crawler.settings.get('USER_AGENT_OS')
        cls.navigator = crawler.settings.get('USER_AGENT_NAVIGATOR')
        cls.device_type = crawler.settings.get('USER_AGENT_DEVICE_TYPE')

        user_agent = generate_user_agent(os=cls.os, navigator=cls.navigator, device_type=cls.device_type)
        o = cls(user_agent)
        crawler.signals.connect(o.spider_opened, signal=signals.spider_opened)
        return o

    def spider_opened(self, spider):
        self.user_agent = getattr(spider, 'user_agent', self.user_agent)

    def process_request(self, request, spider):
        if self.mode == 'per':
            request.headers['User-Agent'] = generate_user_agent(os=self.os, navigator=self.navigator,
                                                                device_type=self.device_type)
        elif self.user_agent:
            request.headers.setdefault(b'User-Agent', self.user_agent)



import pymongo
from twisted.internet.threads import deferToThread


class MongoDBPipeline(object):
    def __init__(self, mongodb_uri, mongodb_database):
        self.mongodb_uri = mongodb_uri
        self.mongodb_database = mongodb_database

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            mongodb_uri=crawler.settings.get('MONGODB_URI'),
            mongodb_database=crawler.settings.get('MONGODB_DATABASE')
        )

    def open_spider(self, spider):
        self.client = pymongo.MongoClient(self.mongodb_uri)
        self.database = self.client[self.mongodb_database]

    def _process_item(self, item, spider):
        allowed_spiders = item.mongodb_spiders
        allowed_collections = item.mongodb_collections
        if allowed_spiders and spider.name in allowed_spiders:
            for allowed_collection in allowed_collections:
                if item.unique_key and item.unique_identity:
                    condition = {item.unique_key: item.get(item.unique_key), 'unique_identity': item.unique_identity}
                    self.database[allowed_collection].update_one(condition, {'$set': dict(item)}, True)
                else:
                    self.database[allowed_collection].insert_one(dict(item))

        return item

    def close_spider(self, spider):
        self.client.close()

    def process_item(self, item, spider):
        return deferToThread(self._process_item, item, spider)



import copy
import json
from collections import UserDict
from typing import Any


class DictPath(UserDict):
    """
        DicPath is a normal dict but you can extract and inject the complex nested dict only using Path.
    """

    # FIX wrong behavior of standard library
    # <a>.pop(<b>, None) returns key error, if <b> not in <a>.
    # This fixes it.
    def pop(self, key, *args):
        return self.data.pop(key, *args)

    def __init__(self, dictionary: dict = {}, deep_copy: bool = False, delimiter='.'):
        """
            Accept python dict or DictPath.
        """
        if isinstance(dictionary, DictPath):
            self.data = dictionary.data
        elif isinstance(dictionary, dict):
            self.data = dictionary
        else:
            raise Exception("dictionary must be a dict")
        if deep_copy:
            self.data = copy.deepcopy(self.data)

        self.delimiter = delimiter

    @property
    def dict(self) -> dict:
        """
            Returns a reference to the dict.
        """
        return self.data

    @property
    def deepcopy(self):
        """
            Create a complete copy of a DictPath
        """
        return DictPath(self, deep_copy=True)

    def __repr__(self) -> str:
        """
            Returns a pretty indented json string representation.
        """
        dump = json.dumps(self.data, indent=2, sort_keys=True)
        return f"DictPath({dump})"

    def clean_path(self, path):
        path = path[1:] if path.startswith(self.delimiter) else path
        return path.split(self.delimiter)

    def get(self, path: str):
        """
            Get the value of the dictionary at the given path
            dict.get("foo/bar/foo1/bar1") is like calling dict["foo"]["bar"]["foo1"]["bar1"].
            Invalid paths return None without error.
        """
        path = self.clean_path(path)
        if path == "":
            return self
        current = self.data
        for attr in path:
            if not isinstance(current, dict):
                raise Exception(f"Your path is not a path of dicts (value at key {attr} is of type {type(current)})")
            if attr not in current:
                return None
            current = current[attr]
        if isinstance(current, dict):
            return DictPath(current)
        return current

    def set(self, path: str, value=None):
        """
            Set the value of the dictionary at the given path
            dict.set("foo/bar/foo1/bar1", 'bar') is like calling dict["foo"]["bar"]["foo1"]["bar1"] = "bar".
            If a path does not exist, it will be created.
            Empty path will do nothing.
        """
        path = self.clean_path(path)
        current = self.data
        last_path_attr = path.pop()
        for attr in path:
            if not isinstance(current, dict):
                raise Exception("Can't set the key of a non-dict")
            current.setdefault(attr, {})
            current = current[attr]
        if isinstance(value, DictPath):
            current[last_path_attr] = value.data
        else:
            current[last_path_attr] = value

    def __getitem__(self, path) -> Any:
        """ Subscript for <DictPath>.get() """
        # If DictPath["key1"], then path="key1"
        # DictPath["key1", "key2"], then path=tuple("key1", "key2")

        path = self.delimiter.join(list(path)) if isinstance(path, tuple) else path
        return self.get(path)

    def __setitem__(self, path, value):
        """ Subscript for <DictPath>.get() and <DictPath>.apply_at_path() """
        path = self.delimiter.join(list(path)) if isinstance(path, tuple) else path
        self.set(path, value=value)


def extract_dict(dictionary, path, delimiter='.'):
    path = path[1:] if path.startswith(delimiter) else path
    paths = path.split(delimiter)
    active_dict = dictionary
    for p in paths:
        if active_dict.get(p) != None:
            active_dict = active_dict.get(p)
        else:
            return None
    return active_dict


def inject_dict(dictionary, path, value, delimiter='.'):
    path = path[1:] if path.startswith(delimiter) else path
    paths = path.split(delimiter)
    path_len = len(paths)
    _active_dict = dictionary
    for i, p in enumerate(paths):
        if i == path_len - 1:
            _active_dict[p] = value
        else:
            if _active_dict.get(p) == None:
                _active_dict[p] = {}
            _active_dict = _active_dict.get(p)
    return dictionary



import redis
from scrapy import signals
from scrapy.exceptions import IgnoreRequest
from scrapy.utils.request import request_fingerprint


class PersistFilterMiddleware:
    def __init__(self):
        self.client: redis.Redis = None
        self.__redis_pool: redis.ConnectionPool = None

    @classmethod
    def from_crawler(cls, crawler):
        cls.redis_host = crawler.settings.get('PERSIST_FILTER_REDIS_HOST')
        cls.redis_port = crawler.settings.get('PERSIST_FILTER_REDIS_PORT')
        cls.redis_password = crawler.settings.get('PERSIST_FILTER_REDIS_PASSWORD')
        cls.persist_filter_name = crawler.settings.get('PERSIST_FILTER_NAME')

        o = cls()
        crawler.signals.connect(o.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(o.spider_closed, signal=signals.spider_closed)
        return o

    def spider_opened(self, spider):
        self.__redis_pool = redis.ConnectionPool(
            host=self.redis_host, port=self.redis_port, password=self.redis_password, decode_responses=True)
        self.client = redis.Redis(connection_pool=self.__redis_pool)

    def spider_closed(self, spider):
        self.__redis_pool.disconnect()

    @staticmethod
    def _get_request_fp(request):
        fp = request_fingerprint(request)
        request.meta['request_fp'] = fp
        return fp

    def process_request(self, request, spider):
        fp = self._get_request_fp(request)
        if request.meta.get('persist_filter'):
            key_name = self.persist_filter_name or spider.name
            fp_exists = self.client.sismember(key_name, fp)
            if fp_exists:
                raise IgnoreRequest("Skipping URL. Already scrapped or in pipeline.")
            else:
                self.client.sadd(key_name, fp)
                return None



import os.path
import io

from scrapy_helper import __version__
from setuptools import setup, find_packages

# Import the README and use it as the long-description.
cwd = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(cwd, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# 依赖
with io.open("requirements.txt", 'r') as f:
    install_requires = f.read().split(os.linesep)

setup(
    name='scrapy-helper',
    version=__version__,
    description='scrapy helper',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    author='Zhou Ping',
    author_email='231409@qq.com',
    keywords=['scrapy', 'spider', 'helper'],
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Framework :: Scrapy',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.7',
    install_requires=install_requires,
)



import scrapy


class FastItem(scrapy.Item):
    def __getitem__(self, key):
        return self._values.get(key, None)

    def __setitem__(self, key, value):
        self._values[key] = value

    def __setattr__(self, key, value):
        object.__setattr__(self, key, value)

    def set(self, fields):
        """
        用来动态添加自定义属性
        """
        for field in fields:
            self.__setattr__(field, scrapy.Field())
            self.fields[field] = scrapy.Field()



import logging
import re
from urllib.parse import parse_qs, urljoin, urlparse

from tldextract import tldextract

log = logging.getLogger(__name__)

_STRICT_DATE_REGEX_PREFIX = r'(?<=\W)'
DATE_REGEX = r'([\./\-_]{0,1}(19|20)\d{2})[\./\-_]{0,1}(([0-3]{0,1}[0-9][\./\-_])|(\w{3,5}[\./\-_]))([0-3]{0,1}[0-9][\./\-]{0,1})?'
STRICT_DATE_REGEX = _STRICT_DATE_REGEX_PREFIX + DATE_REGEX

ALLOWED_TYPES = ['html', 'htm', 'md', 'rst', 'aspx', 'jsp', 'rhtml', 'cgi',
                 'xhtml', 'jhtml', 'asp', 'shtml']

GOOD_PATHS = ['story', 'article', 'feature', 'featured', 'slides',
              'slideshow', 'gallery', 'news', 'video', 'media',
              'v', 'radio', 'press']

BAD_CHUNKS = ['careers', 'contact', 'about', 'faq', 'terms', 'privacy',
              'advert', 'preferences', 'feedback', 'info', 'browse', 'howto',
              'account', 'subscribe', 'donate', 'shop', 'admin']

BAD_DOMAINS = ['amazon', 'doubleclick', 'twitter']


def url_to_filetype(abs_url):
    """
    Input a URL and output the filetype of the file
    specified by the url. Returns None for no filetype.
    'http://blahblah/images/car.jpg' -> 'jpg'
    'http://yahoo.com'               -> None
    """
    path = urlparse(abs_url).path
    # Eliminate the trailing '/', we are extracting the file
    if path.endswith('/'):
        path = path[:-1]
    path_chunks = [x for x in path.split('/') if len(x) > 0]
    last_chunk = path_chunks[-1].split('.')  # last chunk == file usually
    if len(last_chunk) < 2:
        return None
    file_type = last_chunk[-1]
    # Assume that file extension is maximum 5 characters long
    if len(file_type) <= 5 or file_type.lower() in ALLOWED_TYPES:
        return file_type.lower()
    return None


def redirect_back(url, source_domain):
    """
    Some sites like Pinterest have api's that cause news
    args to direct to their site with the real news url as a
    GET param. This method catches that and returns our param.
    """
    parse_data = urlparse(url)
    domain = parse_data.netloc
    query = parse_data.query

    # If our url is even from a remotely similar domain or
    # sub domain, we don't need to redirect.
    if source_domain in domain or domain in source_domain:
        return url

    query_item = parse_qs(query)
    if query_item.get('url'):
        # log.debug('caught redirect %s into %s' % (url, query_item['url'][0]))
        return query_item['url'][0]

    return url


def prepare_url(url, source_url=None):
    """
    Operations that purify a url, removes arguments,
    redirects, and merges relatives with absolutes.
    """
    try:
        if source_url is not None:
            source_domain = urlparse(source_url).netloc
            proper_url = urljoin(source_url, url)
            proper_url = redirect_back(proper_url, source_domain)
            # proper_url = remove_args(proper_url)
        else:
            # proper_url = remove_args(url)
            proper_url = url
    except ValueError as e:
        log.critical('url %s failed on err %s' % (url, str(e)))
        proper_url = ''

    return proper_url


def valid_news_url(url, verbose=False, test=False):
    """
    Is this URL a valid news-article url?

    Perform a regex check on an absolute url.

    First, perform a few basic checks like making sure the format of the url
    is right, (scheme, domain, tld).

    Second, make sure that the url isn't some static resource, check the
    file type.

    Then, search of a YYYY/MM/DD pattern in the url. News sites
    love to use this pattern, this is a very safe bet.

    Separators can be [\.-/_]. Years can be 2 or 4 digits, must
    have proper digits 1900-2099. Months and days can be
    ambiguous 2 digit numbers, one is even optional, some sites are
    liberal with their formatting also matches snippets of GET
    queries with keywords inside them. ex: asdf.php?topic_id=blahlbah
    We permit alphanumeric, _ and -.

    Our next check makes sure that a keyword is within one of the
    separators in a url (subdomain or early path separator).
    cnn.com/story/blah-blah-blah would pass due to "story".

    We filter out articles in this stage by aggressively checking to
    see if any resemblance of the source& domain's name or tld is
    present within the article title. If it is, that's bad. It must
    be a company link, like 'cnn is hiring new interns'.

    We also filter out articles with a subdomain or first degree path
    on a registered bad keyword.
    """
    # If we are testing this method in the testing suite, we actually
    # need to preprocess the url like we do in the article's constructor!
    if test:
        url = prepare_url(url)

    # 11 chars is shortest valid url length, eg: http://x.co
    if url is None or len(url) < 11:
        if verbose: print('\t%s rejected because len of url is less than 11' % url)
        return False

    r1 = ('mailto:' in url)  # TODO not sure if these rules are redundant
    r2 = ('http://' not in url) and ('https://' not in url)

    if r1 or r2:
        if verbose: print('\t%s rejected because len of url structure' % url)
        return False

    path = urlparse(url).path

    # input url is not in valid form (scheme, netloc, tld)
    if not path.startswith('/'):
        return False

    # the '/' which may exist at the end of the url provides us no information
    if path.endswith('/'):
        path = path[:-1]

    # '/story/cnn/blahblah/index.html' --> ['story', 'cnn', 'blahblah', 'index.html']
    path_chunks = [x for x in path.split('/') if len(x) > 0]

    # siphon out the file type. eg: .html, .htm, .md
    if len(path_chunks) > 0:
        file_type = url_to_filetype(url)

        # if the file type is a media type, reject instantly
        if file_type and file_type not in ALLOWED_TYPES:
            if verbose: print('\t%s rejected due to bad filetype' % url)
            return False

        last_chunk = path_chunks[-1].split('.')
        # the file type is not of use to use anymore, remove from url
        if len(last_chunk) > 1:
            path_chunks[-1] = last_chunk[-2]

    # Index gives us no information
    if 'index' in path_chunks:
        path_chunks.remove('index')

    # extract the tld (top level domain)
    tld_dat = tldextract.extract(url)
    subd = tld_dat.subdomain
    tld = tld_dat.domain.lower()

    url_slug = path_chunks[-1] if path_chunks else ''

    if tld in BAD_DOMAINS:
        if verbose: print('%s caught for a bad tld' % url)
        return False

    if len(path_chunks) == 0:
        dash_count, underscore_count = 0, 0
    else:
        dash_count = url_slug.count('-')
        underscore_count = url_slug.count('_')

    # If the url has a news slug title
    if url_slug and (dash_count > 4 or underscore_count > 4):

        if dash_count >= underscore_count:
            if tld not in [x.lower() for x in url_slug.split('-')]:
                if verbose: print('%s verified for being a slug' % url)
                return True

        if underscore_count > dash_count:
            if tld not in [x.lower() for x in url_slug.split('_')]:
                if verbose: print('%s verified for being a slug' % url)
                return True

    # There must be at least 2 subpaths
    if len(path_chunks) <= 1:
        if verbose: print('%s caught for path chunks too small' % url)
        return False

    # Check for subdomain & path red flags
    # Eg: http://cnn.com/careers.html or careers.cnn.com --> BAD
    for b in BAD_CHUNKS:
        if b in path_chunks or b == subd:
            if verbose: print('%s caught for bad chunks' % url)
            return False

    match_date = re.search(DATE_REGEX, url)

    # if we caught the verified date above, it's an article
    if match_date is not None:
        if verbose: print('%s verified for date' % url)
        return True

    for GOOD in GOOD_PATHS:
        if GOOD.lower() in [p.lower() for p in path_chunks]:
            if verbose: print('%s verified for good path' % url)
            return True

    if verbose: print('%s caught for default false' % url)
    return False


def get_domain(abs_url, **kwargs):
    """
    returns a url's domain, this method exists to
    encapsulate all url code into this file
    """
    if abs_url is None:
        return None
    return urlparse(abs_url, **kwargs).netloc


def get_scheme(abs_url, **kwargs):
    """
    """
    if abs_url is None:
        return None
    return urlparse(abs_url, **kwargs).scheme


def get_path(abs_url, **kwargs):
    """
    """
    if abs_url is None:
        return None
    return urlparse(abs_url, **kwargs).path



# -*- coding: utf-8 -*-
from .videos import Video

VIDEOS_TAGS = ['iframe', 'embed', 'object', 'video']
VIDEO_PROVIDERS = ['youtube', 'vimeo', 'dailymotion', 'kewego']


class VideoExtractor(object):
    """Extracts a list of video from Article top node
    """
    def __init__(self, config, top_node):
        self.config = config
        self.parser = self.config.get_parser()
        self.top_node = top_node
        self.candidates = []
        self.movies = []

    def get_embed_code(self, node):
        return "".join([
            line.strip()
            for line in self.parser.nodeToString(node).splitlines()])

    def get_embed_type(self, node):
        return self.parser.getTag(node)

    def get_width(self, node):
        return self.parser.getAttribute(node, 'width')

    def get_height(self, node):
        return self.parser.getAttribute(node, 'height')

    def get_src(self, node):
        return self.parser.getAttribute(node, 'src')

    def get_provider(self, src):
        if src:
            for provider in VIDEO_PROVIDERS:
                if provider in src:
                    return provider
        return None

    def get_video(self, node):
        """Create a video object from a video embed
        """
        video = Video()
        video.embed_code = self.get_embed_code(node)
        video.embed_type = self.get_embed_type(node)
        video.width = self.get_width(node)
        video.height = self.get_height(node)
        video.src = self.get_src(node)
        video.provider = self.get_provider(video.src)
        return video

    def get_iframe_tag(self, node):
        return self.get_video(node)

    def get_video_tag(self, node):
        """Extract html video tags
        """
        return Video()

    def get_embed_tag(self, node):
        # embed node may have an object node as parent
        # in this case we want to retrieve the object node
        # instead of the embed
        parent = self.parser.getParent(node)
        if parent is not None:
            parent_tag = self.parser.getTag(parent)
            if parent_tag == 'object':
                return self.get_object_tag(node)
        return self.get_video(node)

    def get_object_tag(self, node):
        # test if object tag has en embed child
        # in this case we want to remove the embed from
        # the candidate list to avoid parsing it twice
        child_embed_tag = self.parser.getElementsByTag(node, 'embed')
        if child_embed_tag and child_embed_tag[0] in self.candidates:
            self.candidates.remove(child_embed_tag[0])

        # get the object source
        # if we don't have a src node don't coninue
        src_node = self.parser.getElementsByTag(
            node, tag="param", attr="name", value="movie")
        if not src_node:
            return None

        src = self.parser.getAttribute(src_node[0], "value")

        # check provider
        provider = self.get_provider(src)
        if not provider:
            return None

        video = self.get_video(node)
        video.provider = provider
        video.src = src
        return video

    def get_videos(self):
        self.candidates = self.parser.getElementsByTags(
            self.top_node, VIDEOS_TAGS)
        # loop all candidates
        # and check if src attribute belongs to a video provider
        for candidate in self.candidates:
            tag = self.parser.getTag(candidate)
            attr = "get_%s_tag" % tag
            if hasattr(self, attr):
                movie = getattr(self, attr)(candidate)
                if movie is not None and movie.provider is not None:
                    self.movies.append(movie)

        return list(self.movies)
        # append movies list to article
        # self.article.movies = list(self.movies)



from .downloader.user_agent import UserAgentMiddleware
from .downloader.cookies import CookiesMiddleware
from .downloader.proxy import ProxyMiddleware
from .downloader.persist_filter import PersistFilterMiddleware
from .downloader.pyppeteer_middleware import PyppeteerMiddleware



import json
import re

import lxml.html
import lxml.html.clean


def str2list(x, transformer=None):
    """
    convert to list
    :param transformer:
    :param x:
    :return:
    """
    if x is None or isinstance(x, list):
        return x
    try:
        data = json.loads(x)
        if not transformer:
            def transformer(t): return t
        data = list(map(lambda y: transformer(y), data))
        return data
    except:
        return []


def str2dict(v):
    """
    convert str to dict data
    :param v:
    :return:
    """
    try:
        return json.loads(v)
    except:
        return {}


def str2body(v):
    """
    convert str to json data or keep original string
    :param v:
    :return:
    """
    try:
        return json.loads(v)
    except:
        return v


def inner_trim(value):
    if isinstance(value, str):
        # remove tab and white space
        value = re.sub(re.compile(r'[\s\t]+'), ' ', value)
        value = ''.join(value.splitlines())
        return value.strip()
    return ''


def clean_html(node):
    # cleaner = lxml.html.clean.Cleaner()
    # cleaner.javascript = True
    # cleaner.style = True
    # cleaner.allow_tags = [
    #     'a', 'span', 'p', 'br', 'strong', 'b',
    #     'em', 'i', 'tt', 'code', 'pre', 'blockquote', 'img', 'h1',
    #     'h2', 'h3', 'h4', 'h5', 'h6',
    #     'ul', 'ol', 'li', 'dl', 'dt', 'dd']
    # cleaner.remove_unknown_tags = False
    cleaner = lxml.html.clean.Cleaner()
    cleaner.scripts = True
    cleaner.javascript = True
    cleaner.style = True
    return cleaner.clean_html(node)


def convert_to_html(element):
    if not element:
        return ''

    if isinstance(element, str):
        element = lxml.html.fromstring(element)

    clean_data = clean_html(element)
    data = lxml.html.tostring(clean_data, encoding="utf8", pretty_print=True, method='html').decode("utf-8")
    return data


def convert_to_text(element):
    if not element:
        return ''

    if isinstance(element, str):
        element = lxml.html.fromstring(element)

    text = lxml.html.tostring(element, encoding="utf8", method='text').decode("utf-8")
    return inner_trim(text)



from scrapy.linkextractors.lxmlhtml import LxmlLinkExtractor

from scrapy_helper.core.urls import valid_news_url


class ArticleLinkExtractor(LxmlLinkExtractor):
    pass

    def _link_allowed(self, link):
        allowed = super()._link_allowed(link)
        if allowed:
            return valid_news_url(link.url)

        return allowed



import copy

from .cleaners import DocumentCleaner
from .configuration import Configuration
from .extractors import ContentExtractor
from .outputformatters import OutputFormatter
from .utils import extend_config, get_available_languages
from .videos.extractors import VideoExtractor
from ..core import urls


class ArticleException(Exception):
    pass


class Article(object):
    def __init__(self, url, html, title='', source_url='', config=None, **kwargs):
        """The **kwargs argument may be filled with config values, which
        is added into the config object
        """
        if isinstance(title, Configuration) or \
                isinstance(source_url, Configuration):
            raise ArticleException(
                'Configuration object being passed incorrectly as title or '
                'source_url! Please verify `Article`s __init__() fn.')

        self.config = config or Configuration()
        self.config = extend_config(self.config, kwargs)

        self.extractor = ContentExtractor(self.config)

        if source_url == '':
            scheme = urls.get_scheme(url)
            if scheme is None:
                scheme = 'http'
            source_url = scheme + '://' + urls.get_domain(url)

        if source_url is None or source_url == '':
            raise ArticleException('input url bad format')

        # URL to the main page of the news source which owns this article
        self.source_url = source_url

        self.url = urls.prepare_url(url, self.source_url)

        self.title = title

        # URL of the "best image" to represent this article
        self.top_img = ''

        # stores image provided by metadata
        self.meta_img = ''

        # All image urls in this article
        self.images = []

        # All videos in this article: youtube, vimeo, etc
        self.movies = []

        # Body text from this article
        self.text = ''

        # `meta_keywords` are extracted via parse() from <meta> tags
        self.meta_keywords = []

        # `tags` are also extracted via parse() from <meta> tags
        self.tags = []

        # List of authors who have published the article, via parse()
        self.authors = []

        self.publish_date = ''

        # This article's unchanged and raw HTML
        if html and isinstance(html, bytes):
            html = self.config.get_parser().get_unicode_html(html)
        self.html = html

        # The HTML of this article's main node (most important part)
        self.article_html = ''

        # Meta description field in the HTML source
        self.meta_description = ""

        # Meta language field in HTML source
        self.meta_lang = ""

        # Meta favicon field in HTML source
        self.meta_favicon = ""

        # Meta site_name field in HTML source
        self.meta_site_name = ""

        # Meta tags contain a lot of structured data, e.g. OpenGraph
        self.meta_data = {}

        # The canonical link of this article if found in the meta data
        self.canonical_link = ""

        # Holds the top element of the DOM that we determine is a candidate
        # for the main body of the article
        self.top_node = None

        # A deepcopied clone of the above object before heavy parsing
        # operations, useful for users to query data in the
        # "most important part of the page"
        self.clean_top_node = None

        # lxml DOM object generated from HTML
        self.doc = None

        # A deepcopied clone of the above object before undergoing heavy
        # cleaning operations, serves as an API if users need to query the DOM
        self.clean_doc = None

    def parse(self):
        self.doc = self.config.get_parser().fromstring(self.html)
        self.clean_doc = copy.deepcopy(self.doc)

        if self.doc is None:
            # `parse` call failed, return nothing
            return

        document_cleaner = DocumentCleaner(self.config)
        output_formatter = OutputFormatter(self.config)

        title = self.extractor.get_title(self.clean_doc)
        self.set_title(title)

        authors = self.extractor.get_authors(self.clean_doc)
        self.set_authors(authors)

        meta_lang = self.extractor.get_meta_lang(self.clean_doc)
        if meta_lang:
            self.set_meta_language(meta_lang)

            if self.config.use_meta_language:
                self.extractor.update_language(self.meta_lang)
                output_formatter.update_language(self.meta_lang)

        meta_favicon = self.extractor.get_favicon(self.clean_doc)
        self.set_meta_favicon(meta_favicon)

        meta_site_name = self.extractor.get_meta_site_name(self.clean_doc)
        self.set_meta_site_name(meta_site_name)

        meta_description = \
            self.extractor.get_meta_description(self.clean_doc)
        self.set_meta_description(meta_description)

        canonical_link = self.extractor.get_canonical_link(
            self.url, self.clean_doc)
        self.set_canonical_link(canonical_link)

        tags = self.extractor.extract_tags(self.clean_doc)
        self.set_tags(list(tags))

        meta_keywords = self.extractor.get_meta_keywords(
            self.clean_doc)
        self.set_meta_keywords(meta_keywords)

        meta_data = self.extractor.get_meta_data(self.clean_doc)
        self.set_meta_data(meta_data)

        self.publish_date = self.extractor.get_publishing_date(
            self.url,
            self.clean_doc)

        # Before any computations on the body, clean DOM object
        self.doc = document_cleaner.clean(self.doc)

        self.top_node = self.extractor.calculate_best_node(self.doc)
        if self.top_node is not None:
            video_extractor = VideoExtractor(self.config, self.top_node)
            self.set_movies(video_extractor.get_videos())

            self.top_node = self.extractor.post_cleanup(self.top_node)
            self.clean_top_node = copy.deepcopy(self.top_node)

            text, article_html = output_formatter.get_formatted(
                self.top_node)
            self.set_article_html(article_html)
            self.set_text(text)

        self.fetch_images()

    def fetch_images(self):
        if self.clean_doc is not None:
            meta_img_url = self.extractor.get_meta_img_url(
                self.url, self.clean_doc)
            self.set_meta_img(meta_img_url)

            imgs = self.extractor.get_img_urls(self.url, self.clean_doc)
            if self.meta_img:
                imgs.add(self.meta_img)
            self.set_images(list(imgs))

        if self.clean_top_node is not None and not self.has_top_image() and self.images:
            first_img = self.images[0]
            self.set_top_img(first_img)

    def has_top_image(self):
        return self.top_img is not None and self.top_img != ''

    def set_title(self, input_title):
        if input_title:
            self.title = input_title[:self.config.MAX_TITLE]

    def set_text(self, text):
        text = text[:self.config.MAX_TEXT]
        if text:
            self.text = text

    def set_article_html(self, article_html):
        """Sets the HTML of just the article's `top_node`
        """
        if article_html:
            self.article_html = article_html

    def set_meta_img(self, src_url):
        self.meta_img = src_url
        self.set_top_img(src_url)

    def set_top_img(self, src_url):
        self.top_img = src_url

    def set_images(self, images):
        self.images = images

    def set_authors(self, authors):
        """Authors are in ["firstName lastName", "firstName lastName"] format
        """
        if not isinstance(authors, list):
            raise Exception("authors input must be list!")
        if authors:
            self.authors = authors[:self.config.MAX_AUTHORS]

    def set_meta_language(self, meta_lang):
        """Save langauges in their ISO 2-character form
        """
        if meta_lang and len(meta_lang) >= 2 and \
                meta_lang in get_available_languages():
            self.meta_lang = meta_lang[:2]

    def set_meta_keywords(self, meta_keywords):
        """Store the keys in list form
        """
        self.meta_keywords = [k.strip() for k in meta_keywords.split(',')]

    def set_meta_favicon(self, meta_favicon):
        self.meta_favicon = meta_favicon

    def set_meta_site_name(self, meta_site_name):
        self.meta_site_name = meta_site_name

    def set_meta_description(self, meta_description):
        self.meta_description = meta_description

    def set_meta_data(self, meta_data):
        self.meta_data = meta_data

    def set_canonical_link(self, canonical_link):
        self.canonical_link = canonical_link

    def set_tags(self, tags):
        self.tags = tags

    def set_movies(self, movie_objects):
        """Trim video objects into just urls
        """
        movie_urls = [o.src for o in movie_objects if o and o.src]
        self.movies = movie_urls
